﻿(function tracingModule()
{
    "use strict";
    var categories = {
            error: "Error",
            warning: "Warning",
            perf: "Performance",
            info: "Information",
            verbose: "Verbose"
        };
    var registeredListeners = [];
    var listenerFilter = {};
    var recorderFilter = {};
    var maxRecordedTraces = 500;
    var recordedTraces = [];
    var indexNextRecord = 0;
    Object.keys(categories).forEach(function enableTraceCategoryByDefault(categoryKey)
    {
        var category = categories[categoryKey];
        listenerFilter[category] = true;
        recorderFilter[category] = true
    });
    function trace(category, formatText)
    {
        if(BingApp.Utilities.isNullOrUndefined(category))
            throw new BingApp.Classes.ErrorArgumentNullOrUndefined("category");
        var passToListeners = listenerFilter[category];
        var recordInMemory = recorderFilter[category] && maxRecordedTraces > 0;
        if(!passToListeners && !recordInMemory)
            return;
        var formatArgs = Array.prototype.slice.call(arguments,1);
        var traceMessage = BingApp.Utilities.format.apply(null,formatArgs);
        if(passToListeners)
        {
            var countListeners = registeredListeners.length;
            for(var indexListener = 0; indexListener < countListeners; indexListener++)
            {
                var listener = registeredListeners[indexListener];
                listener.log(category,traceMessage)
            }
        }
        if(recordInMemory)
        {
            if(indexNextRecord >= maxRecordedTraces)
                indexNextRecord = 0;
            if(indexNextRecord < recordedTraces.length)
                recordedTraces[indexNextRecord] = traceMessage;
            else
                recordedTraces.push(traceMessage);
            indexNextRecord++
        }
    }
    function traceError(formatText)
    {
        var traceArgs = Array.prototype.slice.call(arguments,0);
        traceArgs.unshift(categories.error);
        trace.apply(null,traceArgs)
    }
    function traceWarning(formatText)
    {
        var traceArgs = Array.prototype.slice.call(arguments,0);
        traceArgs.unshift(categories.warning);
        trace.apply(null,traceArgs)
    }
    function tracePerf(formatText)
    {
        var traceArgs = Array.prototype.slice.call(arguments,0);
        traceArgs.unshift(categories.perf);
        trace.apply(null,traceArgs)
    }
    function traceInfo(formatText)
    {
        var traceArgs = Array.prototype.slice.call(arguments,0);
        traceArgs.unshift(categories.info);
        trace.apply(null,traceArgs)
    }
    function traceVerbose(formatText)
    {
        var traceArgs = Array.prototype.slice.call(arguments,0);
        traceArgs.unshift(categories.verbose);
        trace.apply(null,traceArgs)
    }
    WinJS.Namespace.define("BingApp",{
        TraceCategories: categories,
        trace: trace,
        traceError: traceError,
        traceWarning: traceWarning,
        tracePerf: tracePerf,
        traceInfo: traceInfo,
        traceVerbose: traceVerbose
    });
    function updateTracingSettings(configStore)
    {
        var currentConfig = configStore.getConfig();
        Object.keys(categories).forEach(function enableTraceCategoryAccordingToConfig(categoryKey)
        {
            var category = categories[categoryKey];
            var categorySettings = currentConfig.tracing[categoryKey];
            listenerFilter[category] = categorySettings && categorySettings.enabled;
            recorderFilter[category] = categorySettings && categorySettings.recorded
        });
        var maxRecorded = currentConfig.tracing && currentConfig.tracing.settings && currentConfig.tracing.settings.maxRecorded || 0;
        BingApp.Tracing.setMaxRecordedCount(maxRecorded)
    }
    WinJS.Namespace.define("BingApp.Tracing",{
        setRecorderFilter: function(filter)
        {
            if(!filter)
                throw new BingApp.Classes.ErrorArgumentNullOrUndefined("filter");
            recorderFilter = filter
        },
        getRecorderFilter: function()
        {
            return recorderFilter
        },
        setOutputFilter: function(filter)
        {
            if(!filter)
                throw new BingApp.Classes.ErrorArgumentNullOrUndefined("filter");
            listenerFilter = filter
        },
        getOutputFilter: function()
        {
            return listenerFilter
        },
        addListener: function(listener)
        {
            if(!listener)
                throw new BingApp.Classes.ErrorArgumentNullOrUndefined("listener");
            registeredListeners.push(listener)
        },
        removeListener: function(listener)
        {
            if(!listener)
                throw new BingApp.Classes.ErrorArgumentNullOrUndefined("listener");
            for(var indexListener = registeredListeners.length - 1; indexListener >= 0; indexListener--)
                if(registeredListeners[indexListener] === listener)
                    registeredListeners.splice(indexListener,1)
        },
        setMaxRecordedCount: function(count)
        {
            if(BingApp.Utilities.isNullOrUndefined(count))
                throw new BingApp.Classes.ErrorArgumentNullOrUndefined("count");
            if(count < 0)
                throw new BingApp.Classes.ErrorArgument("count");
            if(maxRecordedTraces !== count)
            {
                maxRecordedTraces = count;
                recordedTraces = BingApp.Tracing.getRecordedTraces(maxRecordedTraces);
                indexNextRecord = recordedTraces.length < maxRecordedTraces ? recordedTraces.length : 0
            }
        },
        getMaxRecordedCount: function()
        {
            return maxRecordedTraces
        },
        getRecordedTraces: function(count)
        {
            var availableTracesCount = recordedTraces.length;
            count = count > availableTracesCount || BingApp.Utilities.isNullOrUndefined(count) ? availableTracesCount : count;
            if(count < 1)
                return[];
            var startIndex = indexNextRecord - count;
            if(startIndex >= 0)
                return recordedTraces.slice(availableTracesCount - count);
            else
                return recordedTraces.slice(availableTracesCount + startIndex).concat(recordedTraces.slice(0,indexNextRecord))
        },
        clearRecordedTraces: function()
        {
            recordedTraces.length = 0;
            indexNextRecord = 0
        },
        init: function(options)
        {
            var configStore = options && options.config || BingApp.locator.config;
            var relay = options && options.eventRelay || BingApp.locator.eventRelay;
            updateTracingSettings(configStore);
            relay.addEventListener(BingApp.Classes.ConfigStore.events.updated,function onConfigUpdated()
            {
                updateTracingSettings(configStore)
            })
        }
    });
    WinJS.Namespace.define("BingApp.Classes",{BaseTraceListener: WinJS.Class.define(function constructor(){},{log: function(category, message)
            {
                var methodName = "BaseTraceListener.log";
                throw new BingApp.Classes.ErrorInvalidOperation(methodName,BingApp.Utilities.format(WinJS.Resources.getString("error_must_override").value,methodName));
            }})});
    WinJS.Namespace.define("BingApp.Classes",{EtwTraceListener: WinJS.Class.derive(BingApp.Classes.BaseTraceListener,function constructor()
        {
            if(!(this instanceof BingApp.Classes.EtwTraceListener))
            {
                BingApp.traceWarning("EtwTraceListener.ctor: Attempted using EtwTraceListener ctor as function; redirecting to use 'new EtwTraceListener()'.");
                return new BingApp.Classes.EtwTraceListener
            }
        },{log: function(category, message)
            {
                message = BingApp.Utilities.format("{0}: {1}",category,message);
                window.msWriteProfilerMark(message)
            }})});
    WinJS.Namespace.define("BingApp.Classes",{ConsoleTraceListener: WinJS.Class.derive(BingApp.Classes.BaseTraceListener,function constructor()
        {
            if(!(this instanceof BingApp.Classes.ConsoleTraceListener))
            {
                BingApp.traceWarning("ConsoleTraceListener.ctor: Attempted using ConsoleTraceListener ctor as function; redirecting to use 'new ConsoleTraceListener()'.");
                return new BingApp.Classes.ConsoleTraceListener
            }
        },{log: function(category, message)
            {
                console.log(message)
            }})});
    if(window.msWriteProfilerMark)
        BingApp.Tracing.addListener(new BingApp.Classes.EtwTraceListener);
    if(window.console && console.log)
        BingApp.Tracing.addListener(new BingApp.Classes.ConsoleTraceListener)
})()