﻿(function aggregateDataProviderModule()
{
    "use strict";
    var AggregateDataProvider = WinJS.Class.define(function constructor(identity, dataProviders)
        {
            if(!(this instanceof BingApp.Classes.AggregateDataProvider))
            {
                BingApp.traceWarning("AggregateDataProvider.ctor: Attempted using AggregateDataProvider ctor as function; redirecting to use 'new AggregateDataProvider()'.");
                return new BingApp.Classes.AggregateDataProvider(identity,dataProviders)
            }
            if(!identity)
                throw new BingApp.Classes.ErrorArgumentNullOrUndefined("identity");
            if(!dataProviders)
                throw new BingApp.Classes.ErrorArgumentNullOrUndefined("dataProviders");
            Object.defineProperties(this,{
                identity: {
                    value: identity,
                    writable: false,
                    enumerable: false,
                    configurable: false
                },
                dataProviderDescriptors: {
                    value: [],
                    writable: false,
                    enumerable: false,
                    configurable: false
                },
                aggregateCount: {
                    writable: true,
                    enumerable: false,
                    configurable: false
                },
                getCountOperation: {
                    writable: true,
                    enumerable: false,
                    configurable: false
                }
            });
            var that = this;
            dataProviders.forEach(function createAndStoreDescriptor(dataProvider)
            {
                that.dataProviderDescriptors.push({dataProvider: dataProvider})
            })
        },{
            getIdentity: function()
            {
                return this.identity
            },
            getCount: function()
            {
                if(this.getCountOperation)
                {
                    BingApp.traceInfo("AggregateDataProvider.getCount: method is called while previous getCount operation did not finish yet; returning promise for getCount operation that is already running.");
                    return this.getCountOperation
                }
                if(!BingApp.Utilities.isNullOrUndefined(this.aggregateCount))
                {
                    BingApp.traceInfo("AggregateDataProvider.getCount: method was called before; returning previously calculated count '{0}'.",this.aggregateCount);
                    return WinJS.Promise.as(this.aggregateCount)
                }
                var that = this;
                this.getCountOperation = new WinJS.Promise(function init(complete, error)
                {
                    var promises = [WinJS.Promise.as(0)];
                    that.dataProviderDescriptors.forEach(function getContainedDataProviderCount(descriptor)
                    {
                        var providerIdentity = descriptor.dataProvider.getIdentity();
                        if(typeof descriptor.count === "number")
                        {
                            BingApp.traceInfo("AggregateDataProvider.getCount: reusing previously calculated count '{0}' for aggregated provider with identity '{1}'.",descriptor.count,providerIdentity);
                            promises.push(WinJS.Promise.as(descriptor.count))
                        }
                        else
                            promises.push(descriptor.dataProvider.getCount().then(function onCompleted(result)
                            {
                                BingApp.traceInfo("AggregateDataProvider.getCount: storing count '{0}' for aggregated provider with identity '{1}'.",result,providerIdentity);
                                descriptor.count = result;
                                return result
                            },function onError(err)
                            {
                                BingApp.traceWarning("AggregateDataProvider.getCount: error returned while trying to determine items count for aggregated provider with identity '{0}'. Error message: '{1}'",providerIdentity,err.message);
                                delete descriptor.count;
                                return WinJS.UI.CountError.noResponse
                            }))
                    });
                    WinJS.Promise.join(promises).then(function onComplete(results)
                    {
                        var aggregatedCount = 0;
                        results.forEach(function combineAggregatedCount(count)
                        {
                            if(typeof count === "number")
                                aggregatedCount += count
                        });
                        complete(aggregatedCount)
                    },error)
                });
                this.getCountOperation.done(function done(result)
                {
                    BingApp.traceInfo("AggregateDataProvider.getCount: calculated aggregate count is '{0}'.",result);
                    that.aggregateCount = result;
                    that.getCountOperation = null
                });
                return this.getCountOperation ? this.getCountOperation : WinJS.Promise.as(this.aggregateCount)
            },
            getItems: function(from, count, processItem)
            {
                if(from < 0)
                {
                    BingApp.traceWarning("AggregateDataProvider.getItems: 'from' index is negative; returning fetch error.");
                    return WinJS.Promise.as(WinJS.UI.FetchError.doesNotExist)
                }
                var that = this;
                if(BingApp.Utilities.isNullOrUndefined(this.aggregateCount))
                {
                    BingApp.traceInfo("AggregateDataProvider.getItems: aggregated count has not been determined yet; wait for getCount completion before continuing with getItems execution.");
                    return this.getCount().then(function completed()
                        {
                            return that.getItems(from,count,processItem)
                        })
                }
                var to = from + count;
                var promises = [];
                var startIndex = 0;
                var providersCount = this.dataProviderDescriptors.length;
                for(var indexProvider = 0; indexProvider < providersCount; indexProvider++)
                {
                    (function getItemsFromContainedDataProvider(descriptor)
                    {
                        if(typeof descriptor.count === "number" && descriptor.count > 0)
                        {
                            var providerIdentity = descriptor.dataProvider.getIdentity();
                            var nextStartIndex = startIndex + descriptor.count;
                            if(from >= startIndex && from < descriptor.count + startIndex)
                            {
                                var localFrom = from - startIndex;
                                var localTo = to - startIndex;
                                if(localTo > descriptor.count)
                                    localTo = descriptor.count;
                                BingApp.traceInfo("AggregateDataProvider.getItems: retrieving items from aggregated provider with identity '{0}' within [{1}..{2}] range.",providerIdentity,localFrom,localTo - 1);
                                if(!processItem && descriptor.items && descriptor.items.length >= localTo)
                                {
                                    BingApp.traceInfo("AggregateDataProvider.getItems: reusing previously retrieved items for aggregated provider with identity '{0}'.",providerIdentity);
                                    promises.push(WinJS.Promise.as({
                                        items: descriptor.items.slice(localFrom,localTo),
                                        totalCount: descriptor.count
                                    }))
                                }
                                else
                                    promises.push(descriptor.dataProvider.getItems(localFrom,localTo - localFrom,processItem).then(function onCompleted(result)
                                    {
                                        if(result.items)
                                            BingApp.traceInfo("AggregateDataProvider.getItems: aggregated provider with identity '{0}' returned '{1}' items.",providerIdentity,result.items.length);
                                        else
                                            BingApp.traceInfo("AggregateDataProvider.getItems: aggregated provider with identity '{0}' did not return items; returned result is '{1}'.",providerIdentity,result);
                                        if(!processItem && result.items)
                                        {
                                            if(!descriptor.items)
                                                descriptor.items = [];
                                            BingApp.traceInfo("AggregateDataProvider.getItems: storing items returned by aggregated provider with identity '{0}' inside cache.",providerIdentity);
                                            var itemsCount = result.items.length;
                                            for(var indexItem = 0; indexItem < itemsCount; indexItem++)
                                                descriptor.items[localFrom + indexItem] = result.items[indexItem]
                                        }
                                        if(!BingApp.Utilities.isNullOrUndefined(result.totalCount) && descriptor.count !== result.totalCount)
                                        {
                                            that.aggregateCount += result.totalCount - descriptor.count;
                                            descriptor.count = result.totalCount;
                                            BingApp.traceInfo("AggregateDataProvider.getItems: aggregated provider with identity '{0}' returned new totalCount value '{1}'. Overall count has been updated to {2}.",providerIdentity,result.totalCount,that.aggregateCount)
                                        }
                                        return result
                                    },function onError(err)
                                    {
                                        BingApp.traceWarning("AggregateDataProvider.getItems: error returned while trying to retrieve items for aggregated provider with identity '{0}'. Error message: '{1}'",providerIdentity,err.message);
                                        delete descriptor.items;
                                        return WinJS.UI.FetchError.doesNotExist
                                    }));
                                from = nextStartIndex
                            }
                            startIndex = nextStartIndex
                        }
                    })(this.dataProviderDescriptors[indexProvider]);
                    if(to <= startIndex)
                        break
                }
                if(promises.length === 0)
                {
                    BingApp.traceInfo("AggregateDataProvider.getItems: none of the aggregated providers return any items. Fetch error will be returned.");
                    return WinJS.Promise.as(WinJS.UI.FetchError.doesNotExist)
                }
                return WinJS.Promise.join(promises).then(function onComplete(results)
                    {
                        var fetchError = true;
                        var aggregateItems = [];
                        results.forEach(function combineAggregatedItems(result)
                        {
                            if(result.items)
                            {
                                fetchError = false;
                                aggregateItems.push.apply(aggregateItems,result.items)
                            }
                        });
                        if(fetchError)
                        {
                            BingApp.traceWarning("AggregateDataProvider.getItems: none of the aggregated providers return any items. Fetch error will be returned.");
                            return WinJS.UI.FetchError.doesNotExist
                        }
                        else
                            return{
                                    items: aggregateItems,
                                    totalCount: that.aggregateCount
                                }
                    })
            },
            invalidate: function(identity)
            {
                BingApp.traceInfo("AggregateDataProvider.invalidate: invalidating internal cached data for {0}.",BingApp.Utilities.isNullOrUndefined(identity) ? "all aggregated providers" : "aggregated provider with identity '" + identity + "'.");
                this.aggregateCount = null;
                this.dataProviderDescriptors.forEach(function invalidateResults(descriptor)
                {
                    if(BingApp.Utilities.isNullOrUndefined(identity) || BingApp.Utilities.areEqualIgnoreCase(identity,descriptor.dataProvider.getIdentity()))
                    {
                        delete descriptor.count;
                        delete descriptor.items
                    }
                })
            }
        });
    WinJS.Namespace.define("BingApp.Classes",{AggregateDataProvider: AggregateDataProvider})
})()