﻿(function listViewModelModule()
{
    "use strict";
    var events = Object.freeze({
            refreshStarting: "refreshStarting",
            refreshFinished: "refreshFinished",
            itemsRemoving: "itemsRemoving",
            itemsRemoved: "itemsRemoved",
            itemsAdding: "itemsAdding",
            itemsAdded: "itemsAdded"
        });
    var ListViewModel = WinJS.Class.derive(BingApp.Classes.EventSource,function constructor(dataProvider, cacheProvider)
        {
            if(!(this instanceof BingApp.Classes.ListViewModel))
            {
                BingApp.traceWarning("ListViewModel.ctor: Attempted using ListViewModel ctor as function; redirecting to use 'new ListViewModel()'.");
                return new BingApp.Classes.ListViewModel(dataProvider,cacheProvider)
            }
            if(!dataProvider)
                throw new BingApp.Classes.ErrorArgumentNullOrUndefined("dataProvider");
            BingApp.Classes.EventSource.call(this);
            Object.defineProperties(this,{
                dataProvider: {
                    value: dataProvider,
                    writable: false,
                    enumerable: false,
                    configurable: false
                },
                cacheProvider: {
                    value: cacheProvider,
                    writable: false,
                    enumerable: false,
                    configurable: false
                },
                list: {
                    value: new WinJS.Binding.List,
                    writable: false,
                    enumerable: false,
                    configurable: false
                },
                refreshOperation: {
                    writable: true,
                    enumerable: false,
                    configurable: false
                }
            });
            if(this.cacheProvider)
            {
                var that = this;
                BingApp.Data.getAllItems(this.cacheProvider).then(function onGetAllItemsComplete(items)
                {
                    var countItemsInCache = items.length;
                    if(countItemsInCache > 0)
                    {
                        BingApp.traceInfo("ListViewModel.ctor: populating list with data stored in cache ('{0} items overall').",countItemsInCache);
                        that.list.push.apply(that.list,items)
                    }
                })
            }
        },{
            getList: function()
            {
                return this.list
            },
            refresh: function(options)
            {
                if(this.refreshOperation)
                {
                    BingApp.traceInfo("ListViewModel.refresh: method is called while previous refresh did not finish yet; returning promise for refresh operation that is already running.");
                    return this.refreshOperation
                }
                BingApp.traceInfo("ListViewModel.refresh: starting operation and raising 'refreshStarting' event.");
                this.fireEvent(events.refreshStarting);
                var that = this;
                var getAllItemsOperation;
                var refreshFinished = false;
                this.refreshOperation = new WinJS.Promise(function init(complete, error)
                {
                    options = options || {};
                    var processItem = options.processItem;
                    var bufferWithCount = options.bufferWithCount || 1;
                    var bufferWithTime = options.bufferWithTime || 0;
                    var collectionCompleted = false;
                    var batchTimeoutExpired = false;
                    function setupBufferTimeout()
                    {
                        batchTimeoutExpired = false;
                        if(bufferWithTime > 0)
                            setTimeout(function onBatchTimeoutExpired()
                            {
                                batchTimeoutExpired = true;
                                if(collectionCompleted)
                                    processBatch()
                            },bufferWithTime);
                        else
                            batchTimeoutExpired = true
                    }
                    setupBufferTimeout();
                    var firstItem = true;
                    var batch = [];
                    function processBatch()
                    {
                        var itemsCountInBatch = batch.length;
                        var continueWithProcessing;
                        if(!batchTimeoutExpired)
                        {
                            BingApp.traceVerbose("ListViewModel.refresh.processBatch: collected '{0}' items in batch but current time buffer of {1} milliseconds has not expired yet. Waiting until timeout expires.",itemsCountInBatch,bufferWithTime);
                            continueWithProcessing = false
                        }
                        else if(collectionCompleted)
                        {
                            BingApp.traceVerbose("ListViewModel.refresh.processBatch: item collection has been completed; need to process last batch of collected items.");
                            continueWithProcessing = true
                        }
                        else if(batch.length >= bufferWithCount && batchTimeoutExpired)
                        {
                            BingApp.traceVerbose("ListViewModel.refresh.processBatch: item collection has not been completed yet but batching conditions have been met for processing items collected so far.");
                            continueWithProcessing = true
                        }
                        else
                        {
                            BingApp.traceVerbose("ListViewModel.refresh.processBatch: collected '{0}' items in batch but current min count buffer of {1} items has not been reached yet. Waiting until necessary number of items is collected.",itemsCountInBatch,bufferWithCount);
                            continueWithProcessing = false
                        }
                        if(continueWithProcessing)
                        {
                            if(firstItem)
                            {
                                firstItem = false;
                                if(that.list.length > 0)
                                {
                                    var existingItems = that.list.map(function getExistingItem(existingItem)
                                        {
                                            return existingItem
                                        });
                                    BingApp.traceVerbose("ListViewModel.refresh.processBatch: first batch of items has arrived; existing content ('{0}' items) and cache will be cleared now. Raising 'itemsRemoving' event.",existingItems.length);
                                    that.fireEvent(events.itemsRemoving,{items: existingItems});
                                    that.list.length = 0;
                                    BingApp.traceVerbose("ListViewModel.refresh.processBatch: existing content and cache has been cleared. Raising 'itemsRemoved' event.");
                                    that.fireEvent(events.itemsRemoved,{items: existingItems})
                                }
                                if(that.cacheProvider)
                                    that.cacheProvider.clear()
                            }
                            if(itemsCountInBatch > 0)
                            {
                                BingApp.traceVerbose("ListViewModel.refresh.processBatch: started processing of '{0}' items in single batch. Raising 'itemsAdding' event.",itemsCountInBatch);
                                if(processItem)
                                    for(var indexItem = 0; indexItem < itemsCountInBatch; indexItem++)
                                    {
                                        var processedItem = processItem(batch[indexItem]);
                                        if(!BingApp.Utilities.isNullOrUndefined(processedItem))
                                            batch[indexItem] = processedItem
                                    }
                                var batchItems = [].concat(batch);
                                that.fireEvent(events.itemsAdding,{items: batchItems});
                                batch.forEach(function appendBatchItem(batchItem)
                                {
                                    that.list.push(batchItem);
                                    BingApp.traceVerbose("ListViewModel.refresh: processed item #{0}",that.list.length)
                                });
                                if(that.cacheProvider)
                                    that.cacheProvider.addItems(batch);
                                BingApp.traceVerbose("ListViewModel.refresh.processBatch: finished processing of '{0}' items in single batch. Raising 'itemsAdded' event.",itemsCountInBatch);
                                that.fireEvent(events.itemsAdded,{items: batchItems});
                                batch.length = 0
                            }
                            if(collectionCompleted)
                            {
                                BingApp.traceInfo("ListViewModel.refresh: operation is completed. Overall collected '{0}' items. Raising 'refreshFinished' event.",that.list.length);
                                that.fireEvent(events.refreshFinished,{canceled: false});
                                refreshFinished = true;
                                that.refreshOperation = null;
                                complete()
                            }
                            else
                                setupBufferTimeout()
                        }
                    }
                    var getAllItemsOptions = Object.create(options);
                    getAllItemsOptions.processItem = function processBatchItem(item)
                    {
                        if(firstItem || !getAllItemsOptions.count || that.list.length < getAllItemsOptions.count)
                        {
                            batch.push(item);
                            processBatch()
                        }
                    };
                    getAllItemsOperation = BingApp.Data.getAllItems(that.dataProvider,getAllItemsOptions);
                    return getAllItemsOperation.then(function onComplete()
                        {
                            collectionCompleted = true;
                            processBatch()
                        },function onError(err)
                        {
                            if(BingApp.Utilities.isPromiseCancellationError(err))
                            {
                                BingApp.traceInfo("ListViewModel.refresh: operation is canceled. Raising 'refreshFinished' event.");
                                that.fireEvent(events.refreshFinished,{canceled: true});
                                refreshFinished = true;
                                that.refreshOperation = null;
                                complete()
                            }
                            else
                                error(err)
                        })
                },function cancel()
                {
                    if(getAllItemsOperation)
                        getAllItemsOperation.cancel()
                });
                if(refreshFinished)
                    this.refreshOperation = null;
                return this.refreshOperation ? this.refreshOperation : WinJS.Promise.as()
            }
        },{events: events});
    WinJS.Namespace.define("BingApp.Classes",{ListViewModel: ListViewModel})
})()