﻿(function listViewModelRefresherModule()
{
    "use strict";
    var localSettings = Windows.Storage.ApplicationData.current.localSettings;
    function readLastRefreshDate(storageKey)
    {
        var refreshData = localSettings.values[storageKey];
        if(refreshData)
            try
            {
                return new Date(JSON.parse(refreshData).lastRefresh)
            }
            catch(error)
            {
                BingApp.traceError("ListViewModelRefresher.readLastRefreshDate: failed to parse refresh data stored in '{0}' inside local settings: '{1}'; error message: '{2}'.",storageKey,refreshData,error.message)
            }
        else
            BingApp.traceInfo("ListViewModelRefresher.readLastRefreshDate: refresh data not yet stored in '{0}' inside local settings",storageKey)
    }
    function storeLastRefreshDate(storageKey, lastRefreshDate)
    {
        localSettings.values[storageKey] = JSON.stringify({lastRefresh: lastRefreshDate})
    }
    function scheduleRefresh(refresher)
    {
        var lastRefreshDate = refresher.lastRefreshDate;
        var waitingPeriod = refresher.interval * 1e3 - (Date.now() - refresher.lastRefreshDate.getTime());
        var operation = (waitingPeriod > 0 ? WinJS.Promise.timeout(waitingPeriod) : WinJS.Promise.as()).then(function onIntervalExpired()
            {
                refresher.lastRefreshDate = new Date;
                return refresher.listViewModel.refresh(refresher.refreshOptions).then(function onRefreshCompleted()
                    {
                        storeLastRefreshDate(refresher.lastRefreshDateStorageKey,refresher.lastRefreshDate);
                        return scheduleRefresh(refresher)
                    },function onRefreshFailed()
                    {
                        refresher.lastRefreshDate = lastRefreshDate;
                        var errorString = "ListViewModelRefresher: refresher.listViewModel.refresh failed";
                        var instObject = {ei: errorString};
                        BingApp.Instrumentation.logError(instObject);
                        BingApp.traceError(errorString)
                    })
            });
        return operation
    }
    var ListViewModelRefresher = WinJS.Class.define(function constructor(listViewModel, options)
        {
            if(!(this instanceof BingApp.Classes.ListViewModelRefresher))
            {
                BingApp.traceWarning("ListViewModelRefresher.ctor: Attempted using ListViewModelRefresher ctor as function; redirecting to use 'new ListViewModelRefresher()'.");
                return new BingApp.Classes.ListViewModelRefresher(listViewModel,options)
            }
            if(!listViewModel)
                throw new BingApp.Classes.ErrorArgumentNullOrUndefined("listViewModel");
            if(!options)
                throw new BingApp.Classes.ErrorArgumentNullOrUndefined("options");
            if(BingApp.Utilities.isNullOrUndefined(options.interval))
                throw new BingApp.Classes.ErrorArgument("options","options.interval must be defined");
            if(options.interval <= 0)
                throw new BingApp.Classes.ErrorArgument("options",BingApp.Utilities.format("options.interval must be positive, specified value is invalid: '{0}'.",options.interval));
            var that = this;
            function handleConfigChange()
            {
                that.onConfigChanged(that)
            }
            Object.defineProperties(this,{
                listViewModel: {
                    value: listViewModel,
                    writable: false,
                    enumerable: false,
                    configurable: false
                },
                interval: {
                    value: options.interval,
                    writable: true,
                    enumerable: false,
                    configurable: false
                },
                lastRefreshDateStorageKey: {
                    value: options.lastRefreshDateStorageKey,
                    writable: false,
                    enumerable: false,
                    configurable: false
                },
                refreshOptions: {
                    value: options.refreshOptions,
                    writable: false,
                    enumerable: false,
                    configurable: false
                },
                onConfigChanged: {
                    value: options.onConfigChanged,
                    writable: false,
                    enumerable: false,
                    configurable: false
                },
                configChangedListener: {
                    value: handleConfigChange,
                    writable: false,
                    enumerable: false,
                    configurable: false
                },
                lastRefreshDate: {
                    value: new Date(0),
                    writable: true,
                    enumerable: false,
                    configurable: false
                },
                currentScheduledRefresh: {
                    writable: true,
                    enumerable: false,
                    configurable: false
                }
            });
            if(this.onConfigChanged)
                BingApp.locator.eventRelay.addEventListener(BingApp.Classes.ConfigStore.events.updated,this.configChangedListener);
            if(this.lastRefreshDateStorageKey)
            {
                var storedDate = readLastRefreshDate(this.lastRefreshDateStorageKey);
                if(storedDate)
                {
                    this.lastRefreshDate = storedDate;
                    BingApp.traceInfo("ListViewModelRefresher.ctor: lastRefreshDate is loaded from storage entry '{0}' and set to '{1}' value.",this.lastRefreshDateStorageKey,storedDate)
                }
                else
                    BingApp.traceInfo("ListViewModelRefresher.ctor: lastRefreshDate was not found in storage under {0} entry; default value is used: '{1}'.",this.lastRefreshDateStorageKey,this.lastRefreshDate)
            }
            else
                BingApp.traceInfo("ListViewModelRefresher.ctor: storage key for lastRefreshDate was not specified; default value is used: '{0}'.",this.lastRefreshDate);
            this.currentScheduledRefresh = scheduleRefresh(this)
        },{
            updateInterval: function(interval)
            {
                if(!this.currentScheduledRefresh)
                    throw new BingApp.Classes.ErrorInvalidOperation("ListViewModelRefresher.updateInterval","Called on disposed ListViewModelRefresher object.");
                if(BingApp.Utilities.isNullOrUndefined(interval))
                    throw new BingApp.Classes.ErrorArgumentNullOrUndefined("interval");
                if(interval <= 0)
                    throw new BingApp.Classes.ErrorArgument("interval",BingApp.Utilities.format("interval must be positive, specified value is invalid: '{0}'.",interval));
                this.currentScheduledRefresh.cancel();
                this.interval = interval;
                this.currentScheduledRefresh = scheduleRefresh(this)
            },
            forceRefresh: function()
            {
                if(!this.currentScheduledRefresh)
                    throw new BingApp.Classes.ErrorInvalidOperation("ListViewModelRefresher.forceRefresh","Called on disposed ListViewModelRefresher object.");
                this.currentScheduledRefresh.cancel();
                this.lastRefreshDate = new Date(0);
                storeLastRefreshDate(this.lastRefreshDateStorageKey,this.lastRefreshDate);
                this.currentScheduledRefresh = scheduleRefresh(this)
            },
            dispose: function()
            {
                if(!this.currentScheduledRefresh)
                    throw new BingApp.Classes.ErrorInvalidOperation("ListViewModelRefresher.dispose","Called on disposed ListViewModelRefresher object.");
                if(this.onConfigChanged)
                    BingApp.locator.eventRelay.removeEventListener(BingApp.Classes.ConfigStore.events.updated,this.configChangedListener);
                this.currentScheduledRefresh.cancel();
                this.currentScheduledRefresh = null
            }
        });
    WinJS.Namespace.define("BingApp.Classes",{ListViewModelRefresher: ListViewModelRefresher})
})()