﻿(function localStorageDataProviderModule()
{
    "use strict";
    var LocalStorageDataProvider = WinJS.Class.define(function constructor(identity, key)
        {
            if(!(this instanceof BingApp.Classes.LocalStorageDataProvider))
            {
                BingApp.traceWarning("LocalStorageDataProvider.ctor: Attempted using LocalStorageDataProvider ctor as function; redirecting to use 'new LocalStorageDataProvider()'.");
                return new BingApp.Classes.LocalStorageDataProvider(identity,key)
            }
            if(!identity)
                throw new BingApp.Classes.ErrorArgumentNullOrUndefined("identity");
            if(!key)
                throw new BingApp.Classes.ErrorArgumentNullOrUndefined("key");
            Object.defineProperties(this,{
                identity: {
                    value: identity,
                    writable: false,
                    enumerable: false,
                    configurable: false
                },
                key: {
                    value: key,
                    writable: false,
                    enumerable: false,
                    configurable: false
                },
                items: {
                    value: [],
                    writable: false,
                    enumerable: false,
                    configurable: false
                },
                itemsCount: {
                    writable: true,
                    enumerable: false,
                    configurable: false
                },
                localStoragePrefix: {
                    value: "LocalStorageDataProvider." + identity,
                    writable: false,
                    enumerable: false,
                    configurable: false
                },
                lastGeneratedId: {
                    value: 0,
                    writable: true,
                    enumerable: false,
                    configurable: false
                },
                localStorageErrorOccurred: {
                    value: false,
                    writable: true,
                    enumerable: false,
                    configurable: false
                }
            });
            var slotForLastGeneratedId = this._getSlotForLastGeneratedId();
            var hasLastGeneratedId = false;
            try
            {
                hasLastGeneratedId = window.localStorage.hasOwnProperty(slotForLastGeneratedId)
            }
            catch(e)
            {
                this.localStorageErrorOccurred = true;
                BingApp.traceError("LocalStorageDataProvider.ctor: window.localStorage failed: '{0}'. Turning off LocalStorageDataProvider.",e.message)
            }
            if(!this.localStorageErrorOccurred && hasLastGeneratedId)
                this.lastGeneratedId = parseInt(window.localStorage.getItem(slotForLastGeneratedId))
        },{
            getIdentity: function()
            {
                return this.identity
            },
            getCount: function()
            {
                var count = this._getCountInternal();
                return WinJS.Promise.as(count)
            },
            getItems: function(from, count, processItem)
            {
                if(from < 0)
                    return WinJS.Promise.as(WinJS.UI.FetchError.doesNotExist);
                var totalCount = this._getCountInternal();
                if(from >= totalCount)
                    return WinJS.Promise.as(WinJS.UI.FetchError.doesNotExist);
                var to = from + count;
                if(to > totalCount)
                    to = totalCount;
                var items = [];
                for(var indexItem = from; indexItem < to; indexItem++)
                {
                    var item = this._getItemInternal(indexItem);
                    if(processItem)
                    {
                        var processedItem = processItem(item);
                        if(!BingApp.Utilities.isNullOrUndefined(processedItem))
                            item = processedItem
                    }
                    items.push(item)
                }
                return WinJS.Promise.as({
                        items: items,
                        totalCount: totalCount
                    })
            },
            addItem: function(item)
            {
                if(!item)
                    throw new BingApp.Classes.ErrorArgumentNullOrUndefined("item");
                this._ensureItemHasKeyValue(item);
                var indexSlot = this._getSlotForIndexFromItemKey(item);
                if(window.localStorage.hasOwnProperty(indexSlot))
                    throw new BingApp.Classes.ErrorArgument("item",WinJS.Resources.getString("error_local_storage_data_provider_item_must_have_unique_key_value").value);
                var index = this._getCountInternal();
                var itemSlot = this._getSlotForItemFromIndex(index);
                window.localStorage.setItem(itemSlot,JSON.stringify(item));
                window.localStorage.setItem(indexSlot,index);
                this.items[index] = item;
                this._setCountInternal(index + 1);
                return WinJS.Promise.as(item)
            },
            updateItem: function(item)
            {
                if(!item)
                    throw new BingApp.Classes.ErrorArgumentNullOrUndefined("item");
                var index = this._getIndexInternal(item);
                var itemSlot = this._getSlotForItemFromIndex(index);
                window.localStorage.setItem(itemSlot,JSON.stringify(item));
                this.items[index] = item;
                return WinJS.Promise.as(item)
            },
            removeItem: function(item)
            {
                if(!item)
                    throw new BingApp.Classes.ErrorArgumentNullOrUndefined("item");
                var indexRemove = this._getIndexInternal(item);
                var removedItem = this._getItemInternal(indexRemove);
                var totalCount = this._getCountInternal();
                var currentSlot = this._getSlotForItemFromIndex(indexRemove);
                for(var indexItem = indexRemove; indexItem < totalCount - 1; indexItem++)
                {
                    var nextSlot = this._getSlotForItemFromIndex(indexItem + 1);
                    var stringifiedItem = window.localStorage.getItem(nextSlot);
                    var existingItem = this._getItemInternal(indexItem + 1);
                    window.localStorage.setItem(currentSlot,stringifiedItem);
                    window.localStorage.setItem(this._getSlotForIndexFromItemKey(existingItem),indexItem);
                    currentSlot = nextSlot
                }
                this.items.splice(indexRemove,1);
                window.localStorage.removeItem(this._getSlotForIndexFromItemKey(removedItem));
                window.localStorage.removeItem(this._getSlotForItemFromIndex(totalCount - 1));
                this._setCountInternal(totalCount - 1);
                return WinJS.Promise.as(removedItem)
            },
            _isDisabled: function()
            {
                return this.localStorageErrorOccurred
            },
            _getCountInternal: function()
            {
                if(BingApp.Utilities.isNullOrUndefined(this.itemsCount))
                    if(this.localStorageErrorOccurred)
                        this.itemsCount = 0;
                    else
                    {
                        var countAsString = window.localStorage.getItem(this._getSlotForTotalCount());
                        if(countAsString)
                            this.itemsCount = parseInt(countAsString);
                        else
                            this.itemsCount = 0
                    }
                return this.itemsCount
            },
            _setCountInternal: function(count)
            {
                this.itemsCount = count;
                window.localStorage.setItem(this._getSlotForTotalCount(),count.toString())
            },
            _getItemInternal: function(index)
            {
                var item = this.items[index];
                if(!item)
                {
                    item = JSON.parse(window.localStorage.getItem(this._getSlotForItemFromIndex(index)));
                    this.items[index] = item
                }
                return item
            },
            _getIndexInternal: function(item)
            {
                var slotForIndex = this._getSlotForIndexFromItemKey(item);
                var indexAsString = window.localStorage.getItem(slotForIndex);
                if(BingApp.Utilities.isNullOrUndefined(indexAsString))
                    throw new BingApp.Classes.ErrorArgument("item",BingApp.Utilities.format(WinJS.Resources.getString("error_local_storage_data_provider_item_must_be_inside_storage").value,item[this.key]));
                return parseInt(indexAsString)
            },
            _ensureItemHasKeyValue: function(item)
            {
                var keyValue = item[this.key];
                if(BingApp.Utilities.isNullOrUndefined(keyValue))
                {
                    item[this.key] = "generatedId." + this.lastGeneratedId;
                    this.lastGeneratedId++;
                    window.localStorage.setItem(this._getSlotForLastGeneratedId(),this.lastGeneratedId.toString())
                }
                else if(typeof keyValue !== "string")
                    throw new BingApp.Classes.ErrorArgument("item",WinJS.Resources.getString("error_local_storage_data_provider_item_must_have_string_key_value").value);
            },
            _getSlotForItemFromIndex: function(index)
            {
                return this.localStoragePrefix + ".items." + index
            },
            _getSlotForIndexFromItemKey: function(item)
            {
                return this.localStoragePrefix + ".keys." + item[this.key]
            },
            _getSlotForTotalCount: function()
            {
                return this.localStoragePrefix + ".count"
            },
            _getSlotForLastGeneratedId: function()
            {
                return this.localStoragePrefix + ".lastGeneratedId"
            }
        });
    WinJS.Namespace.define("BingApp.Classes",{LocalStorageDataProvider: LocalStorageDataProvider})
})()