﻿(function baseAuthenticatorModule()
{
    "use strict";
    var localSettings = Windows.Storage.ApplicationData.current.localSettings;
    var BaseAuthenticator = WinJS.Class.define(function constructor(options)
        {
            BingApp.traceInfo("BaseAuthenticator constructor");
            this._name = options.name;
            this._secureProxy = options.secureProxy || BingApp.locator.secureProxy;
            this._settings = options.settings;
            this._settingsKey = options.settingsKey;
            this._silentSignInPromptType = options.silentSignInPromptType;
            this._frontDoorProviderName = options.frontDoorProviderName;
            if(BingApp.Utilities.isNullOrUndefined(this._silentSignInPromptType))
                throw new BingApp.Classes.ErrorArgumentNullOrUndefined("options.silentSignInPromptType");
            if(!this._settings)
            {
                var value = localSettings.values[this._settingsKey];
                if(value)
                    try
                    {
                        value = JSON.parse(value)
                    }
                    catch(e)
                    {
                        BingApp.traceInfo(this._name + ": BaseAuthenticator constructor: could not parse JSON '{0}' for setting '{1}'.",value,this._settingsKey);
                        value = null
                    }
                if(!value)
                    value = {automaticSignIn: true};
                this._settings = value
            }
        },{
            _name: "BaseAuthenticator",
            _initialExpirySeconds: -1,
            _isAuthenticated: false,
            _silentSignInPromptType: null,
            _authenticationData: {},
            _updateSettings: function()
            {
                localSettings.values[this._settingsKey] = JSON.stringify(this._settings)
            },
            _signInToIFrames: function(parameters, authenticationData)
            {
                var that = this;
                var configuration = that._config.getConfig();
                parameters["cache_bust"] = Date.now();
                parameters["src"] = configuration.authentication.general.sourceName;
                var url = configuration.app.general.secureHost + configuration.authentication.general.tokenSignInPath + BingApp.Utilities.QueryString.serialize(parameters);
                return this._secureProxy.xhr({
                        url: url,
                        type: "GET"
                    }).then(function _signInToIFramesComplete(response)
                    {
                        var success = false;
                        if(response)
                        {
                            var authenticationStatus = JSON.parse(response.responseText);
                            success = authenticationStatus.success;
                            BingApp.traceInfo(that._name + "._signInToIFrames: FrontDoor Authentication success: {0}.",success)
                        }
                        else
                            BingApp.traceInfo(that._name + "._signInToIFrames: FrontDoor Authentication failed: unexpected response: {0}.",response.responseText);
                        if(success)
                            return that._signInCompleted(authenticationData);
                        else
                            return WinJS.Promise.wrapError(new WinJS.ErrorFromName(that._name + " provided a token, but it was not accepted by FrontDoor."))
                    },function _signInToIFramesError(error)
                    {
                        BingApp.traceError(that._name + "._signInToIFrames: sign in error");
                        return WinJS.Promise.wrapError(error)
                    })
            },
            _signInCompleted: function(authenticationData)
            {
                this._isAuthenticated = true;
                this._authenticationData = authenticationData;
                this._settings.automaticSignIn = true;
                this._updateSettings();
                BingApp.traceInfo(this._name + "._signInCompleted: sign in completed");
                return WinJS.Promise.as()
            },
            _signOutOfIFrames: function()
            {
                var that = this;
                return this._secureProxy.getCookie("_SS","SID").then(function _signOutOfIFramesSIDComplete(sid)
                    {
                        var configuration = that._config.getConfig();
                        var signOutUrl = configuration.app.general.secureHost + configuration.authentication.general.tokenSignOutPath + BingApp.Utilities.QueryString.serialize({
                                completion_mode: "image",
                                provider: that._frontDoorProviderName,
                                sig: sid,
                                cache_bust: Date.now(),
                                src: configuration.authentication.general.sourceName
                            });
                        BingApp.traceInfo(that._name + "._signOutOfIFramesSIDComplete: SID crumb received {0}",sid);
                        return that._secureProxy.xhr({
                                url: signOutUrl,
                                type: "GET"
                            })
                    }).then(function _signOutOfIFramesComplete(result)
                    {
                        if(!result || result.status !== 200)
                        {
                            BingApp.traceError(that._name + "._signOutOfIFrames: sign out failed: {0}",result && JSON.stringify(result));
                            return WinJS.Promise.wrapError(new WinJS.ErrorFromName("secure proxy response failed"))
                        }
                        BingApp.traceInfo(that._name + "._signOutOfIFrames: sign out successful: {0}",JSON.stringify(result))
                    },function _signOutOfIFramesError(error)
                    {
                        BingApp.traceError(that._name + "_signOutOfIFrames: sign out failed - error name: {0}, error message: {1}",error.name,error.message);
                        return WinJS.Promise.wrapError(error)
                    })
            },
            _signOutCompleted: function()
            {
                this._isAuthenticated = false;
                this._initialExpirySeconds = -1;
                this._authenticationData = {};
                this._settings.automaticSignIn = false;
                this._updateSettings();
                BingApp.traceInfo(this._name + "._signOutCompleted: sign out completed");
                return WinJS.Promise.as()
            },
            trySilentSignIn: function()
            {
                BingApp.traceInfo("BaseAuthenticator.trySilentSignIn: trySilentSignIn started");
                var that = this;
                if(!("automaticSignIn" in that._settings) || that._settings.automaticSignIn)
                    return this._authenticate(this._silentSignInPromptType).then(function trySilentSignInComplete()
                        {
                            BingApp.traceInfo(that._name + ".trySilentSignIn: silent sign in completed");
                            return WinJS.Promise.as(true)
                        },function trySilentSignInError(error)
                        {
                            BingApp.traceWarning(that._name + ".trySilentSignIn: silent sign in failed, attempting to sign out user from app to clear cookies; name: {0}, message: {1}.",error.name,error.message);
                            return that._signOutOfIFrames().then(function signOutOfIFramesComplete()
                                {
                                    BingApp.traceInfo(that._name + ".trySilentSignIn: sign in failed, successfully signed out user from app to clear cookies");
                                    return that._signOutCompleted()
                                }).then(function signOutOfIFramesCompleteReturnOriginalError()
                                {
                                    return WinJS.Promise.as(false)
                                })
                        });
                else
                {
                    BingApp.traceInfo(that._name + ": Silent sign-in denied, user had previously signed out");
                    return WinJS.Promise.as(false)
                }
            },
            getAuthenticationData: function()
            {
                return this._authenticationData
            },
            getExpiry: function()
            {
                return this._initialExpirySeconds
            },
            isAuthenticated: function()
            {
                return this._isAuthenticated
            },
            signIn: function()
            {
                var methodName = "BaseAuthenticator.signIn";
                throw new BingApp.Classes.ErrorInvalidOperation(methodName,BingApp.Utilities.format(WinJS.Resources.getString("error_must_override").value,methodName));
            },
            signOut: function()
            {
                var methodName = "BaseAuthenticator.signOut";
                throw new BingApp.Classes.ErrorInvalidOperation(methodName,BingApp.Utilities.format(WinJS.Resources.getString("error_must_override").value,methodName));
            }
        });
    WinJS.Namespace.define("BingApp.Classes",{BaseAuthenticator: BaseAuthenticator})
})()