﻿(function configStoreModule()
{
    "use strict";
    var descriptorKey = "descriptor";
    var remoteSettingsKey = "remoteSettings";
    var overrideSettingsKey = "overrideSettings";
    var testOverrideSettingsPath = "/shell/configuration/test.appconfiguration.json";
    var events = Object.freeze({updated: "config:updated"});
    function getCachedSettings(identity, descriptor, type)
    {
        try
        {
            var localSettings = Windows.Storage.ApplicationData.current.localSettings
        }
        catch(e)
        {
            BingApp.traceWarning("ConfigStore.getCachedSettings: error accessing localSettings: " + e.message);
            return
        }
        var fullSettingsKey = identity + "." + type;
        var fullDescriptorKey = fullSettingsKey + "." + descriptorKey;
        try
        {
            var cachedDescriptor = localSettings.values[fullDescriptorKey]
        }
        catch(e)
        {
            BingApp.traceWarning("ConfigStore.getCachedSettings: error reading descriptor from localSettings: " + e.message);
            return
        }
        if(!checkCompatibility(cachedDescriptor,descriptor,type))
            return;
        try
        {
            var jsonSettings = localSettings.values[fullSettingsKey]
        }
        catch(e)
        {
            BingApp.traceWarning("ConfigStore.getCachedSettings: error reading JSON remote settings from localSettings: " + e.message);
            return
        }
        if(!jsonSettings)
            return;
        try
        {
            var settings = JSON.parse(jsonSettings)
        }
        catch(e)
        {
            BingApp.traceWarning("ConfigStore.getCachedSettings: error parsing JSON remote settings: " + e.message);
            return
        }
        if(typeof settings !== "object")
            return;
        return settings
    }
    function checkCompatibility(cachedDescriptor, descriptor, type)
    {
        if(cachedDescriptor === descriptor)
            return true;
        if(type === overrideSettingsKey)
        {
            var oldDescriptor = parseDescriptor(cachedDescriptor);
            var newDescriptor = parseDescriptor(descriptor);
            if(!oldDescriptor || !newDescriptor)
                return false;
            if(oldDescriptor.region !== newDescriptor.region || oldDescriptor.language !== newDescriptor.language)
                return false;
            var versionResult = BingApp.Utilities.compareVersions(newDescriptor.version,oldDescriptor.version);
            if(typeof versionResult === "undefined" || versionResult < 0)
                return false;
            return versionResult >= 0
        }
        else
            return false
    }
    function parseDescriptor(descriptor)
    {
        if(!descriptor || !descriptor.split)
            return null;
        var items = descriptor.split(",");
        if(items.length !== 3)
            return null;
        var version = items[0];
        var region = items[1];
        var language = items[2];
        return{
                version: version,
                region: region,
                language: language
            }
    }
    function setCachedSettings(identity, descriptor, type, jsonSettings)
    {
        try
        {
            var localSettings = Windows.Storage.ApplicationData.current.localSettings
        }
        catch(e)
        {
            BingApp.traceWarning("ConfigStore.setCachedSettings: error accessing localSettings: " + e.message);
            return
        }
        var fullSettingsKey = identity + "." + type;
        var fullDescriptorKey = fullSettingsKey + "." + descriptorKey;
        try
        {
            localSettings.values[fullDescriptorKey] = descriptor
        }
        catch(e)
        {
            BingApp.traceWarning("ConfigStore.setCachedSettings: error writing descriptor to localSettings: " + e.message);
            return
        }
        try
        {
            localSettings.values[fullSettingsKey] = jsonSettings
        }
        catch(e)
        {
            BingApp.traceWarning("ConfigStore.setCachedSettings: error writing JSON remote settings to localSettings: " + e.message);
            return
        }
    }
    function isRegionVariantSatisfied(region, regionVariant)
    {
        return BingApp.Utilities.areEqualIgnoreCase(regionVariant,region)
    }
    function isLanguageVariantSatisfied(language, languageVariant)
    {
        language = language.toLowerCase();
        languageVariant = languageVariant.toLowerCase();
        if(languageVariant === language)
            return true;
        if(language.length > languageVariant.length && language[languageVariant.length] === "-" && language.indexOf(languageVariant) === 0)
            return true;
        return false
    }
    function isConstraintSatisfied(constraint, region, language)
    {
        var regionSatisfied = constraint.region === undefined || isRegionVariantSatisfied(region,constraint.region);
        var languageSatisfied = constraint.language === undefined || isLanguageVariantSatisfied(language,constraint.language);
        return regionSatisfied && languageSatisfied
    }
    function resolveSetting(setting, region, language)
    {
        if(Array.isArray(setting))
        {
            for(var i = 0; i < setting.length; i++)
                if(isConstraintSatisfied(setting[i],region,language))
                    return setting[i].value
        }
        else
            return setting;
        BingApp.traceError("ConfigStore.resolveSetting: encountered setting with no default value");
        return null
    }
    function resolveSection(section, region, language)
    {
        var resolvedSection = {};
        for(var setting in section)
            resolvedSection[setting] = resolveSetting(section[setting],region,language);
        return resolvedSection
    }
    function resolveFeature(feature, region, language)
    {
        var resolvedFeature = {};
        for(var section in feature)
            resolvedFeature[section] = resolveSection(feature[section],region,language);
        return resolvedFeature
    }
    function resolveConfig(localSettings, region, language)
    {
        var resolvedConfig = {};
        for(var feature in localSettings)
            resolvedConfig[feature] = resolveFeature(localSettings[feature],region,language);
        return resolvedConfig
    }
    function applyOverrideSettings(config, overrideSettings)
    {
        for(var i in overrideSettings)
        {
            var overrideFeature = overrideSettings[i];
            if(typeof overrideFeature === "object")
                for(var j in overrideFeature)
                {
                    var overrideSection = overrideFeature[j];
                    if(typeof overrideSection === "object")
                        for(var k in overrideSection)
                        {
                            var overrideSetting = overrideSection[k];
                            var overrideSettingType = typeof overrideSetting;
                            if(overrideSettingType === "string" || overrideSettingType === "number" || overrideSettingType === "boolean")
                                if(config[i] && config[i][j] && config[i][j][k] !== undefined)
                                    if(typeof config[i][j][k] === overrideSettingType)
                                        config[i][j][k] = overrideSetting;
                                    else
                                        BingApp.traceWarning("ConfigStore.applyOverrideSettings: override setting has different type from local setting '{0}.{1}.{2}'",i,j,k);
                                else
                                    BingApp.traceWarning("ConfigStore.applyOverrideSettings: override setting tried to override non-existent local setting '{0}.{1}.{2}'",i,j,k)
                        }
                }
        }
    }
    function configHasChanged(oldConfig, newConfig)
    {
        for(var i in newConfig)
            for(var j in newConfig[i])
                for(var k in newConfig[i][j])
                    if(!oldConfig[i] || !oldConfig[i][j] || oldConfig[i][j][k] !== newConfig[i][j][k])
                        return true;
        return false
    }
    var ConfigStore = WinJS.Class.define(function constructor(identity, localSettings, options)
        {
            this._identity = identity;
            this._localSettings = localSettings;
            this._region = options && options.region || Windows.System.UserProfile.GlobalizationPreferences.homeGeographicRegion;
            this._language = options && options.language || Windows.System.UserProfile.GlobalizationPreferences.languages[0];
            this._descriptor = options && options.descriptor || ConfigStore._getDescriptor();
            this.reload(false)
        },{
            _getCachedSettings: function(type)
            {
                return getCachedSettings(this._identity,this._descriptor,type)
            },
            _setCachedSettings: function(type, jsonSettings)
            {
                setCachedSettings(this._identity,this._descriptor,type,jsonSettings)
            },
            reload: function(signalIfChanged)
            {
                var newConfig = resolveConfig(this._localSettings,this._region,this._language);
                var remoteSettings = this._getCachedSettings(remoteSettingsKey);
                if(remoteSettings)
                    applyOverrideSettings(newConfig,remoteSettings);
                var overrideSettings = this._getCachedSettings(overrideSettingsKey);
                if(overrideSettings)
                    applyOverrideSettings(newConfig,overrideSettings);
                var testOverrideSettings = this._testOverrideSettings;
                if(testOverrideSettings)
                    applyOverrideSettings(newConfig,testOverrideSettings);
                if(!this._config || configHasChanged(this._config,newConfig))
                {
                    this._config = newConfig;
                    if(signalIfChanged)
                        BingApp.locator.eventRelay.fireEvent(events.updated)
                }
            },
            getConfig: function()
            {
                return this._config
            },
            _remoteUpdate: function()
            {
                var that = this;
                var options = {url: "/win8app/config" + BingApp.locator.env.getQueryStringWithIFrameParams()};
                BingApp.locator.proxy.xhr(options).done(function remoteUpdateCompleted(response)
                {
                    var jsonRemoteSettings = response.responseText;
                    that._setCachedSettings(remoteSettingsKey,jsonRemoteSettings);
                    that.reload(true)
                },function remoteUpdateFailed(err)
                {
                    var errorStatus = err && err.status;
                    var errorMessage = BingApp.Utilities.format("ConfigStore._remoteUpdate: error retrieving remote settings: status = {0}",errorStatus);
                    var instObject = {
                            sc: errorStatus,
                            url: options.url,
                            ei: errorMessage
                        };
                    BingApp.Instrumentation.logError(instObject);
                    BingApp.traceWarning(errorMessage)
                })
            },
            forceRemoteUpdate: function()
            {
                var remoteUpdateClosure = BingApp.Utilities.createClosure(this,this._remoteUpdate);
                setImmediate(remoteUpdateClosure)
            },
            startPeriodicRemoteUpdate: function(updateInterval)
            {
                if(this._remoteUpdateTimer)
                    clearInterval(this._remoteUpdateTimer);
                var remoteUpdateClosure = BingApp.Utilities.createClosure(this,this._remoteUpdate);
                this._remoteUpdateTimer = setInterval(remoteUpdateClosure,updateInterval)
            },
            loadTestOverrideSettings: function()
            {
                var that = this;
                WinJS.xhr({url: testOverrideSettingsPath}).done(function getTestConfigSucceeded(response)
                {
                    if(response.responseText)
                    {
                        try
                        {
                            that._testOverrideSettings = JSON.parse(response.responseText)
                        }
                        catch(e)
                        {
                            BingApp.traceError("ConfigStore.loadTestOverrideSettings: error parsing test override settings as JSON: {0}",e.description);
                            return
                        }
                        that.reload(true)
                    }
                },function getTestConfigFailed(err)
                {
                    var description = err && err.description;
                    BingApp.traceWarning("ConfigStore.loadTestOverrideSettings: XHR request to test override settings file failed with error: {0}",description)
                })
            },
            setOverrideSetting: function(feature, section, setting, value)
            {
                if(BingApp.Utilities.isNullOrUndefined(value))
                {
                    BingApp.traceError("ConfigStore.setOverrideSetting: invalid value {0} passed to override setting {1} in section {2} for feature {3}.",value,setting,section,feature);
                    return
                }
                var overrideSettings = this._getCachedSettings(overrideSettingsKey) || {};
                overrideSettings[feature] || (overrideSettings[feature] = {});
                overrideSettings[feature][section] || (overrideSettings[feature][section] = {});
                overrideSettings[feature][section][setting] = value;
                try
                {
                    var jsonOverrideSettings = JSON.stringify(overrideSettings)
                }
                catch(e)
                {
                    BingApp.traceError("ConfigStore.setOverrideSetting: failed to convert local overrides object into JSON");
                    return
                }
                this._setCachedSettings(overrideSettingsKey,jsonOverrideSettings);
                this.reload(true)
            },
            clearOverrideSetting: function(feature, section, setting)
            {
                var overrideSettings = this._getCachedSettings(overrideSettingsKey) || {};
                if(overrideSettings[feature] && overrideSettings[feature][section])
                    delete overrideSettings[feature][section][setting];
                try
                {
                    var jsonOverrideSettings = JSON.stringify(overrideSettings)
                }
                catch(e)
                {
                    BingApp.traceError("ConfigStore.clearOverrideSetting: failed to convert local overrides object into JSON");
                    return
                }
                this._setCachedSettings(overrideSettingsKey,jsonOverrideSettings);
                this.reload(true)
            }
        },{
            _getDescriptor: function()
            {
                var version = Windows.ApplicationModel.Package.current.id.version;
                var fullVersion = [version.major,version.minor,version.build,version.revision].join(".");
                var region = Windows.System.UserProfile.GlobalizationPreferences.homeGeographicRegion;
                var language = Windows.System.UserProfile.GlobalizationPreferences.languages[0];
                return[fullVersion,region,language].join(",")
            },
            clearCache: function(identity)
            {
                try
                {
                    var localSettings = Windows.Storage.ApplicationData.current.localSettings
                }
                catch(e)
                {
                    BingApp.traceWarning("ConfigStore.clearCache: error accessing localSettings: " + e.message);
                    return
                }
                delete localSettings.values[identity + "." + remoteSettingsKey];
                delete localSettings.values[identity + "." + remoteSettingsKey + "." + descriptorKey];
                delete localSettings.values[identity + "." + overrideSettingsKey];
                delete localSettings.values[identity + "." + overrideSettingsKey + "." + descriptorKey]
            },
            events: events
        });
    WinJS.Namespace.define("BingApp.Classes",{ConfigStore: ConfigStore})
})()