﻿(function eventRelayModule()
{
    "use strict";
    var appGlobalSource = "Win8App";
    var allowedDomainsForTargetOrigin = /^(.+\.)?(bing|bing-int|staging-bing-int|working-bing-int|dev[147]-bing-int|prodmirror1-bing-int)\.com\/?$/i;
    var events = Object.freeze({
            subscribe: "eventrelay:subscribe",
            unsubscribe: "eventrelay:unsubscribe"
        });
    function constructor()
    {
        if(!(this instanceof BingApp.Classes.EventRelay))
        {
            BingApp.traceWarning("EventRelay.ctor: Attempted using EventRelay ctor as function; redirecting to use 'new EventRelay()'.");
            return new BingApp.Classes.EventRelay
        }
        Object.defineProperties(this,{
            eventRegistry: {
                value: {},
                writable: false,
                enumerable: false,
                configurable: false
            },
            iframeRegistry: {
                value: {},
                writable: false,
                enumerable: false,
                configurable: false
            }
        })
    }
    function registerIframe(id, iframe, targetOrigin)
    {
        if(!iframe)
            throw new BingApp.Classes.ErrorArgumentNullOrUndefined("iframe");
        if(targetOrigin)
        {
            var targetOriginForValidation = targetOrigin;
            var indexQueryStart = targetOrigin.indexOf("?");
            if(indexQueryStart !== -1)
                targetOriginForValidation = targetOrigin.substring(0,indexQueryStart);
            else
            {
                var indexFragmentStart = targetOrigin.indexOf("#");
                if(indexFragmentStart !== -1)
                    targetOriginForValidation = targetOrigin.substring(0,indexFragmentStart)
            }
            if(!BingApp.Utilities.areEqualIgnoreCase(targetOriginForValidation,BingApp.locator.config.getConfig().app.general.host) && !BingApp.Utilities.areEqualIgnoreCase(targetOriginForValidation,window.location.protocol + "//" + window.location.host))
            {
                var host = new Windows.Foundation.Uri(targetOriginForValidation).host;
                if(!allowedDomainsForTargetOrigin.test(host))
                {
                    var errorMessage = "EventRelay.registerIframe: Security error attempting to set targetOrigin to non-whitelisted domain: " + targetOrigin;
                    var instObject = {
                            url: host,
                            ei: errorMessage
                        };
                    BingApp.Instrumentation.logError(instObject);
                    throw new BingApp.Classes.ErrorArgument("targetOrigin",BingApp.Utilities.format("Given targetOrigin '{0}' is not allowed for security reasons.",targetOrigin));
                }
            }
        }
        var iframeInfo = this.getRegisteredIframeInfo(id);
        if(iframeInfo)
            if(iframeInfo.iframe !== iframe)
                throw new BingApp.Classes.ErrorArgument("id",BingApp.Utilities.format("Cannot override existing registered provider '{0}'",id));
            else
            {
                iframeInfo.targetOrigin = targetOrigin;
                BingApp.traceInfo("EventRelay.registerIframe: updated targetOrigin to '{0}' for registered subscriber '{1}'",targetOrigin,id)
            }
        else
        {
            this.iframeRegistry[id] = {
                iframe: iframe,
                targetOrigin: targetOrigin
            };
            BingApp.traceInfo("EventRelay.registerIframe: registered subscriber '{0}'",id)
        }
    }
    function unregisterIframe(id)
    {
        var iframeInfo = this.getRegisteredIframeInfo(id);
        if(iframeInfo)
        {
            delete this.iframeRegistry[id];
            var registry = this.eventRegistry;
            Object.keys(registry).forEach(function removeSubscription(eventName)
            {
                var listeners = registry[eventName];
                for(var indexListener = listeners.length - 1; indexListener >= 0; indexListener--)
                {
                    var listener = listeners[indexListener];
                    if(listener.iframeId === id)
                        listeners.splice(indexListener,1)
                }
            });
            BingApp.traceInfo("EventRelay.unregisterIframe: unregistered iframe '{0}' and removed all its subscriptions to events.",id)
        }
        else
            BingApp.traceError("EventRelay.unregisterIframe: there is no registered iframe '{0}'.",id)
    }
    function getRegisteredIframeInfo(id)
    {
        if(!id)
            throw new BingApp.Classes.ErrorArgumentNullOrUndefined("id");
        return this.iframeRegistry[id]
    }
    function fireEvent(eventName, args, source)
    {
        if(BingApp.Utilities.isNullOrUndefined(eventName))
            throw new BingApp.Classes.ErrorArgumentNullOrUndefined("eventName");
        var listeners = this.eventRegistry[eventName];
        if(listeners && listeners.length > 0)
        {
            var listenersCopy = [].concat(listeners);
            var defaultTargetOrigin = BingApp.locator.config.getConfig().app.general.host;
            var messageName = eventName;
            var messageSource = typeof source === "string" ? source : appGlobalSource;
            var messageValue = args;
            var message;
            var that = this;
            listenersCopy.forEach(function fireEventForRegisteredListener(listener)
            {
                if(BingApp.Utilities.isNullOrUndefined(listener.source) || source === listener.source)
                {
                    if(listener.callback)
                        listener.callback(args,source);
                    var id = listener.iframeId;
                    if(!BingApp.Utilities.isNullOrUndefined(id))
                    {
                        var iframeInfo = that.getRegisteredIframeInfo(id);
                        if(!iframeInfo)
                            BingApp.traceError("EventRelay.fireEvent: ignoring iframe listener with '{0}' id. There is no registration information with that id.",id);
                        else
                        {
                            if(!message)
                                try
                                {
                                    message = JSON.stringify({
                                        name: messageName,
                                        value: messageValue,
                                        from: messageSource
                                    })
                                }
                                catch(error)
                                {
                                    BingApp.traceError("EventRelay.fireEvent: error while trying to stringify event '{0}'; replacing args with error information. Error name: '{1}'. Error message: '{2}'.",eventName,error.name,error.message);
                                    message = JSON.stringify({
                                        name: messageName,
                                        error: {
                                            name: error.name,
                                            message: error.message
                                        },
                                        from: messageSource
                                    })
                                }
                            var messageTargetOrigin = iframeInfo.targetOrigin || defaultTargetOrigin;
                            BingApp.traceInfo("EventRelay.fireEvent: relaying from app to web: '{0}'. Target iframe id is '{1}' and origin is '{2}'.",message,id,messageTargetOrigin);
                            var iframe = iframeInfo.iframe;
                            if(iframe.contentWindow)
                                iframe.contentWindow.postMessage(message,messageTargetOrigin);
                            else
                                BingApp.traceError("EventRelay.fireEvent: listener iframe '{0}' has a null or undefined contentWindow.",id)
                        }
                    }
                }
            })
        }
    }
    function addEventListener(eventName, listener)
    {
        if(BingApp.Utilities.isNullOrUndefined(eventName))
            throw new BingApp.Classes.ErrorArgumentNullOrUndefined("eventName");
        listener = normalizeEventListener(listener);
        if(typeof listener.callback !== "function" && BingApp.Utilities.isNullOrUndefined(listener.iframeId))
            throw new BingApp.Classes.ErrorArgument("listener",WinJS.Resources.getString("error_event_listener_must_have_callback_or_iframe_id").value);
        var listeners = this.eventRegistry[eventName];
        if(!listeners)
        {
            listeners = [];
            this.eventRegistry[eventName] = listeners
        }
        listeners.push(listener)
    }
    function removeEventListener(eventName, listener)
    {
        if(BingApp.Utilities.isNullOrUndefined(eventName))
            throw new BingApp.Classes.ErrorArgumentNullOrUndefined("eventName");
        listener = normalizeEventListener(listener);
        var listeners = this.eventRegistry[eventName];
        if(listeners)
            for(var indexListener = 0; indexListener < listeners.length; indexListener++)
            {
                var existingListener = listeners[indexListener];
                if(areEqualEventListeners(listener,existingListener))
                {
                    listeners.splice(indexListener,1);
                    return
                }
            }
        BingApp.traceWarning("EventRelay.removeEventListener: handler was not found for '{0}' event.",eventName)
    }
    function isRegisteredEventListener(eventName, listener)
    {
        if(listener)
        {
            var listeners = this.eventRegistry[eventName];
            if(listeners)
            {
                listener = normalizeEventListener(listener);
                for(var indexListener = 0; indexListener < listeners.length; indexListener++)
                {
                    var existingListener = listeners[indexListener];
                    if(areEqualEventListeners(listener,existingListener))
                        return true
                }
            }
        }
        return false
    }
    function normalizeEventListener(listener)
    {
        if(!listener)
            throw new BingApp.Classes.ErrorArgumentNullOrUndefined("listener");
        if(typeof listener === "function")
            listener = {callback: listener};
        else if(typeof listener === "string")
            listener = {iframeId: listener};
        return listener
    }
    function areEqualEventListeners(listener1, listener2)
    {
        return listener1 === listener2 || listener1 && listener2 && listener1.source === listener2.source && listener1.callback === listener2.callback && listener1.iframeId === listener2.iframeId
    }
    var instanceMembers = {
            fireEvent: fireEvent,
            addEventListener: addEventListener,
            removeEventListener: removeEventListener,
            isRegisteredEventListener: isRegisteredEventListener,
            registerIframe: registerIframe,
            unregisterIframe: unregisterIframe,
            getRegisteredIframeInfo: getRegisteredIframeInfo
        };
    var staticMembers = {
            normalizeEventListener: normalizeEventListener,
            areEqualEventListeners: areEqualEventListeners,
            events: events
        };
    WinJS.Namespace.define("BingApp.Classes",{EventRelay: WinJS.Class.define(constructor,instanceMembers,staticMembers)})
})()