﻿(function geoLocationModule()
{
    "use strict";
    var positionStatus = Object.freeze({
            ready: 0,
            initializing: 1,
            noData: 2,
            disabled: 3,
            notInitialized: 4,
            notAvailable: 5
        });
    var events = Object.freeze({
            positionChanged: "positionchanged",
            statusChanged: "statuschanged"
        });
    var GeoLocationService = WinJS.Class.define(function constructor(options)
        {
            this._proxy = options && options.proxy || BingApp.locator.proxy;
            this._config = options && options.config || BingApp.locator.config;
            this._geolocator = options && options.geolocator || new Windows.Devices.Geolocation.Geolocator;
            this._geolocationConfig = this._config.getConfig().geolocation
        },{
            init: function()
            {
                if(this._geolocationConfig.general.enabled)
                {
                    this._geolocator.reportInterval = this._geolocationConfig.general.geolocatorReportInterval;
                    this._geolocator.movementThreshold = this._geolocationConfig.general.geolocatorMovementThreshold;
                    this._lastUpdatedLocation = null;
                    var onStatusChangedClosure = this._onStatusChanged.bind(this);
                    this._geolocator.addEventListener(events.statusChanged,onStatusChangedClosure);
                    this._onPositionChangedClosure = this._onPositionChanged.bind(this);
                    BingApp.traceInfo("Geolocation: init: _updateGeolocatorPerStatus invoked: geolocatorStatus: {0}",this._geolocator.locationStatus);
                    this._updateGeolocatorPerStatus(this._geolocator.locationStatus)
                }
            },
            _onPositionChanged: function(event)
            {
                if(event)
                    this._updatePosition(event.position);
                else
                    BingApp.traceWarning("Geolocation: _onPositionChanged: event object is Null")
            },
            _updatePosition: function(position)
            {
                var coordinatePrecision = this._geolocationConfig.general.coordinatePrecision;
                var latitude = position && position.coordinate && position.coordinate.latitude && position.coordinate.latitude.toFixed(coordinatePrecision);
                var longitude = position && position.coordinate && position.coordinate.longitude && position.coordinate.longitude.toFixed(coordinatePrecision);
                this._currentLocation = {
                    latitude: latitude,
                    longitude: longitude,
                    timeStamp: Date.now()
                };
                if(!this._currentLocation.latitude || !this._currentLocation.longitude)
                    return;
                if(this._lastUpdatedLocation)
                {
                    if(this._lastUpdatedLocation.timeStamp)
                    {
                        var timeSinceLastUpdate = this._currentLocation.timeStamp - this._lastUpdatedLocation.timeStamp;
                        if(timeSinceLastUpdate < this._geolocator.reportInterval)
                            return
                    }
                    if(this._currentLocation.latitude === this._lastUpdatedLocation.latitude && this._currentLocation.longitude === this._lastUpdatedLocation.longitude)
                        return
                }
                var that = this;
                this._performReverseGeocoding().then(function onSuccess()
                {
                    that._proxy.setLocation(that._currentLocation).done(function onComplete()
                    {
                        that._lastUpdatedLocation = that._currentLocation;
                        BingApp.traceInfo("Geolocation:_updatePosition: user location has been updated." + "(lat:{0}, long:{1}, Disp: {2}, {3})",that._currentLocation.latitude,that._currentLocation.longitude,that._currentLocation.city,that._currentLocation.state)
                    },function onError(err)
                    {
                        BingApp.traceError("Geolocation:_updatePosition: _proxy.setLocation returned error promise: {0}",err && err.message)
                    })
                },function onError(err)
                {
                    BingApp.traceError("Geolocation:_updatePosition: performReverseGeocoding returned error promise: {0}",err && err.message)
                })
            },
            _performReverseGeocoding: function()
            {
                var that = this;
                return new WinJS.Promise(function reverseGeocoding(onReverseGeocodingPromiseSuccess, onReverseGeocodingPromiseError)
                    {
                        if(that._geolocationConfig.general.reverseGeocodingEnabled)
                        {
                            if(that._currentLocation && that._currentLocation.latitude && that._currentLocation.longitude)
                            {
                                var mapsReverseGeocodingServiceUrl = BingApp.Utilities.format(that._geolocationConfig.general.mapsReverseGeocodingServiceUrl,that._currentLocation.latitude,that._currentLocation.longitude);
                                var detail = {url: mapsReverseGeocodingServiceUrl};
                                WinJS.xhr(detail).done(function onComplete(response)
                                {
                                    if(response)
                                    {
                                        try
                                        {
                                            var responseObject = JSON.parse(response.responseText)
                                        }
                                        catch(err)
                                        {
                                            BingApp.traceError("Geolocation:_performReverseGeocoding: JSON.parse on response from Maps service failed. '{0}'.",err.message)
                                        }
                                        var address = responseObject && responseObject.resourceSets[0] && responseObject.resourceSets[0].resources[0] && responseObject.resourceSets[0].resources[0].address;
                                        if(address)
                                            if(address.locality && address.adminDistrict)
                                            {
                                                that._currentLocation.city = address.locality;
                                                that._currentLocation.state = address.adminDistrict;
                                                that._currentLocation.zipcode = address.postalCode;
                                                onReverseGeocodingPromiseSuccess();
                                                return
                                            }
                                            else
                                                BingApp.traceWarning("Geolocation:_performReverseGeocoding: " + "Error: Address does not contain either locality or adminDistrict attribute." + "User location might not be supported fully by reverseGeocoding service." + "Webservice responseText:" + response.responseText);
                                        else
                                            BingApp.traceWarning("Geolocation:_performReverseGeocoding: Error: Null value for address object." + "Webservice responseText:" + response.responseText)
                                    }
                                    else
                                        BingApp.traceError("Geolocation:_performReverseGeocoding: Error: xhr returned null response");
                                    onReverseGeocodingPromiseError(new WinJS.ErrorFromName("Geolocation ReverseGeocoding Error","_performReverseGeocoding:" + " Error: Execution should not reach here."))
                                },function onError(error)
                                {
                                    BingApp.traceError("Geolocation:_performReverseGeocoding: xhr returned error promise.: " + "status: {0} {1}.",error && error.status,error && error.statusText);
                                    onReverseGeocodingPromiseError(new WinJS.ErrorFromName("Geolocation ReverseGeocoding Error","_performReverseGeocoding: xhr returned error promise"))
                                })
                            }
                        }
                        else
                            onReverseGeocodingPromiseSuccess()
                    })
            },
            _onStatusChanged: function(eventData)
            {
                if(eventData)
                {
                    BingApp.traceInfo("Geolocation: _onStatusChanged: statusChanged event fired: geolocatorStatus: {0}",eventData.status);
                    this._updateGeolocatorPerStatus(eventData.status)
                }
                else
                    BingApp.traceError("Geolocation: _onStatusChanged: eventData object is Null")
            },
            _updateGeolocatorPerStatus: function(geolocatorStatus)
            {
                var that = this;
                if(geolocatorStatus === positionStatus.ready)
                    this._geolocator.getGeopositionAsync().then(function onSuccess(position)
                    {
                        that._updatePosition(position);
                        that._geolocator.addEventListener(events.positionChanged,that._onPositionChangedClosure);
                        BingApp.traceInfo("Geolocation: _updateGeolocatorPerStatus : geolocation service is ready.")
                    },function onError(err)
                    {
                        BingApp.traceError("Geolocation: _updateGeolocatorPerStatus: getGeopositionAsync promise returned with error: '{0}'.",err && err.message);
                        that._geolocator.addEventListener(events.positionChanged,that._onPositionChangedClosure)
                    });
                else if(geolocatorStatus !== positionStatus.initializing && geolocatorStatus !== positionStatus.notInitialized)
                {
                    BingApp.traceInfo("Geolocation: _updateGeolocatorPerStatus: geolocation service is 'unavailable'. " + "User location won't be used. geolocatorStatus = {0}",geolocatorStatus);
                    this._lastUpdatedLocation = null;
                    this._proxy.clearLocation().done(function onComplete()
                    {
                        BingApp.traceInfo("Geolocation: _updateGeolocatorPerStatus: location info cleared.")
                    },function onError(err)
                    {
                        BingApp.traceError("Geolocation: _updateGeolocatorPerStatus: clearLocation promise returned with error: '{0}'.",err && err.message)
                    })
                }
            }
        },{
            positionStatus: positionStatus,
            events: events
        });
    WinJS.Namespace.define("BingApp.Classes",{GeoLocationService: GeoLocationService})
})()