﻿(function viewManagementModule()
{
    "use strict";
    var LoadViewOperation = WinJS.Class.derive(BingApp.Classes.AsyncOperation,function constructor()
        {
            BingApp.Classes.AsyncOperation.call(this,"LoadViewOperation")
        },{
            matchOptions: function(options1, options2)
            {
                var location1 = this._getViewLocationFromOptions(options1);
                var location2 = this._getViewLocationFromOptions(options2);
                var host1 = this._getViewHostFromOptions(options1);
                var host2 = this._getViewHostFromOptions(options2);
                return location1 === location2 && host1 === host2
            },
            handleStart: function(options)
            {
                var viewLocation = this._getViewLocationFromOptions(options);
                BingApp.tracePerf("LoadViewOperation.handleStart: started loading view from '{0}' location.",viewLocation)
            },
            handleExecute: function(options)
            {
                var viewLocation = this._getViewLocationFromOptions(options);
                var host = this._getViewHostFromOptions(options);
                return new WinJS.Promise(function init(complete, error)
                    {
                        WinJS.Utilities.empty(host);
                        WinJS.UI.ListView.triggerDispose();
                        WinJS.UI.Pages.render(viewLocation,host).done(complete,error)
                    })
            },
            handleComplete: function(options, result, canceled, error)
            {
                var viewLocation = this._getViewLocationFromOptions(options);
                if(error)
                    BingApp.tracePerf("LoadViewOperation.handleComplete: failed to load view from '{0}' location; error - '{1}'.",viewLocation,error.message);
                else if(canceled)
                    BingApp.tracePerf("LoadViewOperation.handleComplete: cancelled loading of view from '{0}' location.",viewLocation);
                else
                    BingApp.tracePerf("LoadViewOperation.handleComplete: successfully loaded view from '{0}' location.",viewLocation)
            },
            _getViewLocationFromOptions: function(options)
            {
                if(BingApp.Utilities.isNullOrUndefined(options))
                    throw new BingApp.Classes.ErrorArgumentNullOrUndefined("options");
                var viewLocation = options.location;
                if(BingApp.Utilities.isNullOrUndefined(viewLocation))
                    throw new BingApp.Classes.ErrorArgument("options");
                return viewLocation
            },
            _getViewHostFromOptions: function(options)
            {
                if(BingApp.Utilities.isNullOrUndefined(options))
                    throw new BingApp.Classes.ErrorArgumentNullOrUndefined("options");
                var host = options.host;
                if(!host)
                    throw new BingApp.Classes.ErrorArgument("options");
                return host
            }
        });
    var ViewManager = WinJS.Class.derive(BingApp.Classes.EventSource,function constructor(registrySourceLocation)
        {
            if(!(this instanceof BingApp.Classes.ViewManager))
            {
                BingApp.traceWarning("ViewManager.ctor: Attempted using ViewManager ctor as function; redirecting to use 'new ViewManager()'.");
                return new BingApp.Classes.ViewManager(registrySourceLocation)
            }
            BingApp.Classes.EventSource.call(this);
            Object.defineProperties(this,{
                viewState: {
                    value: [],
                    writable: false,
                    enumerable: false,
                    configurable: false
                },
                loadViewOperation: {
                    value: new LoadViewOperation,
                    writable: false,
                    enumerable: false,
                    configurable: false
                },
                currentLayerIndex: {
                    value: -1,
                    writable: true,
                    enumerable: false,
                    configurable: true
                },
                currentLoadingViewOperation: {
                    writable: true,
                    enumerable: false,
                    configurable: true
                },
                loadRegistryOperation: {
                    value: new BingApp.Classes.LoadJsonOperation,
                    writable: false,
                    enumerable: false,
                    configurable: false
                },
                registry: {
                    writable: true,
                    enumerable: false,
                    configurable: false
                }
            });
            if(registrySourceLocation)
                this._loadRegistry(registrySourceLocation)
        },{
            reset: function()
            {
                BingApp.traceInfo("ViewManager.reset: setting up view manager to handle view loading for new navigation operation.");
                this._cancelLoading();
                this.currentLayerIndex = -1
            },
            load: function(viewId)
            {
                if(BingApp.Utilities.isNullOrUndefined(viewId))
                    throw new BingApp.Classes.ErrorArgumentNullOrUndefined("viewId");
                var viewInfo;
                var navigationManager;
                this.currentLayerIndex++;
                BingApp.traceInfo("ViewManager.load: started loading '{0}' into the layer slot with index {1}.",viewId,this.currentLayerIndex);
                if(this.viewState.length > this.currentLayerIndex)
                {
                    viewInfo = this.viewState[this.currentLayerIndex];
                    if(viewInfo.viewId === viewId)
                    {
                        var view = viewInfo.view;
                        if(view)
                        {
                            navigationManager = BingApp.locator.navigationManager;
                            var navigationUri = navigationManager.getCurrentUri();
                            var navigationOptions = navigationManager.getCurrentOptions();
                            BingApp.traceInfo("ViewManager.load: '{0}' is already loaded. View controller will be notified to handle navigation Uri: {1}.",viewId,navigationUri.absoluteUri);
                            var controller = view.getController();
                            if(controller)
                                controller.setNavigationUri(navigationUri,navigationOptions);
                            return WinJS.Promise.as(view)
                        }
                        else
                        {
                            if(this.currentLayerIndex !== this.viewState.length - 1)
                                return WinJS.Promise.wrapError(new BingApp.Classes.ErrorInvalidOperation("ViewManager.load",BingApp.Utilities.format(WinJS.Resources.getString("error_view_must_be_loaded").value,viewId)));
                            BingApp.traceInfo("ViewManager.load: '{0}' is already being loaded. Returning reference to active loading operation.",viewId);
                            if(!this.loadViewOperation.isRunning() || !this.currentLoadingViewOperation)
                                return WinJS.Promise.wrapError(new BingApp.Classes.ErrorInvalidOperation("ViewManager.load",BingApp.Utilities.format(WinJS.Resources.getString("error_must_be_loading_view").value,viewId)));
                            return this.currentLoadingViewOperation
                        }
                    }
                    else
                        this._unloadRemainingViews()
                }
                viewInfo = {viewId: viewId};
                this.viewState[this.currentLayerIndex] = viewInfo;
                var that = this;
                this.currentLoadingViewOperation = new WinJS.Promise(function init(complete, error)
                {
                    that._ready().done(function onComplete()
                    {
                        var host;
                        if(that.currentLayerIndex === 0)
                        {
                            BingApp.traceInfo("ViewManager.load: '{0}' view will be loaded inside Shell.",viewId);
                            var shell = BingApp.locator.shell;
                            if(shell)
                                host = shell.getHost()
                        }
                        else
                        {
                            var parentViewInfo = that.viewState[that.currentLayerIndex - 1];
                            BingApp.traceInfo("ViewManager.load: '{0}' view will be loaded inside '{1}' view.",viewId,parentViewInfo.viewId);
                            var parentController = parentViewInfo.view.getController();
                            if(parentController)
                                host = parentController.getHost()
                        }
                        if(!host)
                            error(new BingApp.Classes.ErrorInvalidOperation("ViewManager.load",BingApp.Utilities.format(WinJS.Resources.getString("error_view_must_have_host").value,viewId)));
                        var entry = that._findRegistryEntry(viewId);
                        if(!entry || !entry.location)
                            error(new BingApp.Classes.ErrorViewRegistration(viewId));
                        that.loadViewOperation.run({
                            location: entry.location,
                            host: host
                        }).done(function onComplete(viewObject)
                        {
                            if(!viewObject || !(viewObject.getController instanceof Function))
                                error(new BingApp.Classes.ErrorViewLoadingFailure(viewId));
                            var viewController = viewObject.getController();
                            if(!viewController)
                                error(new BingApp.Classes.ErrorViewLoadingFailure(viewId));
                            if(that.currentLayerIndex !== that.viewState.length - 1)
                                error(new BingApp.Classes.ErrorInvalidOperation("ViewManager.load",WinJS.Resources.getString("error_view_state_corrupted").value));
                            var loadedViewInfo = that.viewState[that.currentLayerIndex];
                            if(!loadedViewInfo || !BingApp.Utilities.isNullOrUndefined(loadedViewInfo.view))
                                error(new BingApp.Classes.ErrorInvalidOperation("ViewManager.load",WinJS.Resources.getString("error_view_state_corrupted").value));
                            loadedViewInfo.view = viewObject;
                            navigationManager = BingApp.locator.navigationManager;
                            viewController.setNavigationUri(navigationManager.getCurrentUri(),navigationManager.getCurrentOptions());
                            complete(viewObject)
                        },error)
                    },error)
                },function cancel()
                {
                    that._cancelLoading()
                });
                this.currentLoadingViewOperation.done(function onCurrentLoadingViewOperationComplete(){},function onCurrentLoadingViewOperationError(err)
                {
                    that.viewState.pop();
                    that.currentLayerIndex--
                });
                return this.currentLoadingViewOperation
            },
            setRegistry: function(registry)
            {
                BingApp.traceInfo("ViewManager.setRegistry: view registry is set explicitly.");
                this.registry = registry;
                if(this.loadRegistryOperation.isRunning())
                {
                    BingApp.traceWarning("ViewManager.setRegistry: called while app is in the middle of loading view registry from external source. This loading operation will be cancelled.");
                    this.loadRegistryOperation.cancel()
                }
            },
            getLoadedViews: function()
            {
                var loadedViews = [];
                var viewsCount = this.viewState.length;
                var index;
                if(viewsCount > 0)
                {
                    for(index = 0; index < viewsCount - 1; index++)
                        loadedViews.push(this.viewState[index].view);
                    var lastView = this.viewState[index].view;
                    if(!BingApp.Utilities.isNullOrUndefined(lastView))
                        loadedViews.push(lastView)
                }
                return loadedViews
            },
            findLoadedView: function(viewId)
            {
                var viewsCount = this.viewState.length;
                var index;
                for(index = 0; index < viewsCount; index++)
                {
                    var viewState = this.viewState[index];
                    if(viewState && BingApp.Utilities.areEqualIgnoreCase(viewId,viewState.viewId) && viewState.view)
                        return viewState.view
                }
                return null
            },
            _cancelLoading: function()
            {
                this.loadViewOperation.cancel()
            },
            _unloadRemainingViews: function()
            {
                for(var indexView = this.viewState.length - 1; indexView >= this.currentLayerIndex; indexView--)
                {
                    var viewInfo = this.viewState[indexView];
                    var view = viewInfo.view;
                    if(view)
                    {
                        var controller = view.getController();
                        if(controller)
                            controller.notifyUnloading()
                    }
                    else
                    {
                        if(indexView !== this.viewState.length - 1)
                            throw new BingApp.Classes.ErrorInvalidOperation("ViewManager.load",BingApp.Utilities.format(WinJS.Resources.getString("error_must_be_loading_view").value,viewInfo.viewId));
                        this._cancelLoading()
                    }
                }
                this.viewState.length = this.currentLayerIndex
            },
            _findRegistryEntry: function(viewId)
            {
                if(this.registry)
                {
                    var viewsCount = this.registry.length;
                    for(var indexRegistration = 0; indexRegistration < viewsCount; indexRegistration++)
                    {
                        var entry = this.registry[indexRegistration];
                        if(BingApp.Utilities.areEqualIgnoreCase(entry.viewId,viewId))
                            return entry
                    }
                }
                return null
            },
            _ready: function(callback, async)
            {
                var that = this;
                return new WinJS.Promise(function init(complete, error)
                    {
                        function onRegistryLoaded()
                        {
                            if(callback)
                                try
                                {
                                    callback()
                                }
                                catch(err)
                                {
                                    error(err);
                                    return
                                }
                            complete()
                        }
                        function onLoadingError(loadingErr)
                        {
                            if(that.registry && BingApp.Utilities.isPromiseCancellationError(loadingErr))
                                onRegistryLoaded();
                            else
                                error(loadingErr)
                        }
                        if(that.loadRegistryOperation.isRunning())
                            that.loadRegistryOperation.done(onRegistryLoaded,onLoadingError);
                        else if(async)
                            WinJS.Promise.timeout().done(onRegistryLoaded,onLoadingError);
                        else
                            onRegistryLoaded()
                    })
            },
            _loadRegistry: function(fileLocation)
            {
                var that = this;
                return this.loadRegistryOperation.run({fileLocation: fileLocation}).then(function onLoadRegistryOperationComplete(result)
                    {
                        that.registry = result
                    })
            }
        });
    WinJS.Namespace.define("BingApp.Classes",{ViewManager: ViewManager})
})()